import json
from mitoinstaller.user_install import USER_JSON_PATH
import sys
import pytest
import subprocess
from typing import List
from mitoinstaller.commands import uninstall_pip_packages
import os

@pytest.fixture
def clear_user_json():
    """
    This fixture reads in the original user.json file that exists before these tests are
    run, deletes it, and then recreates it at the end. This allows us to test what happens 
    when the user.json file is in various states of out of date and disrepair.

    It also turns off the sending of logging to make sure we don't generate a huge number
    of new accounts on the frontend.
    """
    saved = False
    if os.path.exists(USER_JSON_PATH):
        saved = True
        with open(USER_JSON_PATH, 'r') as f:
            user_json = json.loads(f.read())

        os.remove(USER_JSON_PATH)

    # Turn off logging
    from mitoinstaller.log_utils import analytics
    analytics.send = False

    yield # All tests in this user module run right here

    if saved:
        with open(USER_JSON_PATH, 'w+') as f:
            f.write(json.dumps(user_json))

@pytest.fixture
def venv(tmpdir):
    """
    A fixture that you can use to test
    various parts of the mitoinstaller package
    in an isolated environment.
    """
    # Before running tests, we uninstall all packages from the main virtual environment
    # that could mess with the running of the tests
    print("Removing packages from venv")
    try:
        uninstall_pip_packages('jupyterlab')
    except:
        pass
    try:
        uninstall_pip_packages('mitosheet2')
    except:
        pass
    try:
        uninstall_pip_packages('mitosheet')
    except:
        pass
    try:
        uninstall_pip_packages('mitosheet3')
    except:
        pass

    venv = VirtualEnvironment(tmpdir.strpath)
    yield venv

WINDOWS = os.name == 'nt'

class VirtualEnvironment(object):
    def __init__(self, path):
        self.path = path
        subprocess.check_call([sys.executable, '-m', 'venv', self.path])
        
        self.bin = os.path.join(
            self.path,
            'bin' if not WINDOWS else 'Scripts',
        )
        if not WINDOWS:
            self.python = os.path.join(
                self.path,
                'bin',
                'python3'
            )
        else:
            self.python = os.path.join(
                self.path,
                'Scripts',
                'python.exe'
            )

    def run_python_module_command(self, command: List[str]):
        completed_process = subprocess.run(
            [self.python, '-m'] + command, 
            # NOTE: we do not use the capture_output variable, as this doesn't work before
            # python 3.7, and we want users to be able to install before that
            stdout=subprocess.PIPE, 
            stderr=subprocess.STDOUT,
            # NOTE: we use universal_newlines to get the result back as text, 
            # but we don't use text=True, because we want to work before 3.7 when
            # text was introduced. See here: https://stackoverflow.com/questions/41171791/how-to-suppress-or-capture-the-output-of-subprocess-run
            universal_newlines=True
        )
        stdout = completed_process.stdout if isinstance(completed_process.stdout, str) else ''
        stderr = completed_process.stderr if isinstance(completed_process.stderr, str) else ''
        return stdout, stderr

    def run_python_script(self, script):
        output = subprocess.check_output([self.python, '-c', script]).decode('utf-8')
        return output

    def get_package_version(self, package):
        script = """
import pkg_resources
try:
    print(pkg_resources.get_distribution("{package}").version)
except:
    print('')
        """.format(package=package)
        version = self.run_python_script(script).strip()
        if version == '':
            return None
        return version

    def get_jupyterlab_extension_names(self):
        """
        Helper function that returns a list of installed extensions
        """
        from mitoinstaller.commands import get_extension_names_from_labextension_list_output
        stdout, stderr = self.run_python_module_command(['jupyter', 'labextension', 'list'])
        return get_extension_names_from_labextension_list_output(stdout, stderr)


from copy import deepcopy
import pytest
import json
import os
from mitoinstaller.experiments.experiment_utils import get_new_experiment

from tests.conftest import VirtualEnvironment, clear_user_json

from mitoinstaller.user_install import USER_JSON_DEFAULT, get_current_experiment, get_static_user_id, try_create_user_json_file
from mitoinstaller.user_install import USER_JSON_PATH


def get_mitosheet_version_in_user_json():
    try:
        return json.loads(open(USER_JSON_PATH, 'r').read())['mitosheet_current_version']
    except:
        return None


def test_import_mitosheet_keeps_same_user_id(venv: VirtualEnvironment, clear_user_json):
    venv.run_python_module_command(['pip', 'install', '-r', 'requirements.txt'])    
    # First, check there is no user.json
    static_user_id_installer = get_static_user_id()
    assert static_user_id_installer is None
    venv.run_python_module_command(['mitoinstaller', 'install'])

    # We get the user id created by the installer
    static_user_id_installer = get_static_user_id()
    assert static_user_id_installer is not None
    
    # Then, we import Mito
    venv.run_python_script('import mitosheet')
    
    # And check that the static id has not changed, but the rest
    # of the user.json file has 
    static_user_id_new = get_static_user_id()
    assert static_user_id_installer == static_user_id_new
    mitosheet_version = get_mitosheet_version_in_user_json()
    assert mitosheet_version is not None

def test_installer_does_not_overwrite_static_user_id(venv: VirtualEnvironment, clear_user_json):
    venv.run_python_module_command(['pip', 'install', '-r', 'requirements.txt'])    
    # First, check there is no user.json
    static_user_id_installer = get_static_user_id()
    assert static_user_id_installer is None

    # Then, create an id
    try_create_user_json_file()
    static_user_id_before_installer = get_static_user_id()

    # Then, run the installer
    venv.run_python_module_command(['mitoinstaller', 'install'])

    # We get the user id created by the installer
    static_user_id_installer = get_static_user_id()
    assert static_user_id_installer is not None
    
    # Then, we import Mito
    venv.run_python_script('import mitosheet')
    
    # And check that the static id has not changed, but the rest
    # of the user.json file has 
    static_user_id_new = get_static_user_id()
    assert static_user_id_before_installer == static_user_id_installer
    assert static_user_id_installer == static_user_id_new
    mitosheet_version = get_mitosheet_version_in_user_json()
    assert mitosheet_version is not None


def test_running_installer_twice_does_not_overwrite_static_user_id(venv: VirtualEnvironment, clear_user_json):
    venv.run_python_module_command(['pip', 'install', '-r', 'requirements.txt'])    
    # First, check there is no user.json
    static_user_id_installer = get_static_user_id()
    assert static_user_id_installer is None

    # Then, run the installer
    venv.run_python_module_command(['mitoinstaller', 'install'])
    static_user_id_one = get_static_user_id()

    venv.run_python_module_command(['mitoinstaller', 'install'])
    static_user_id_two = get_static_user_id()

    assert static_user_id_one == static_user_id_two

def test_generates_experiment(venv: VirtualEnvironment, clear_user_json):
    venv.run_python_module_command(['pip', 'install', '-r', 'requirements.txt'])    
    venv.run_python_module_command(['mitoinstaller', 'install'])
    current_experiment = get_current_experiment()
    assert current_experiment is not None

def test_overwrites_old_experiment(venv: VirtualEnvironment, clear_user_json):
    venv.run_python_module_command(['pip', 'install', '-r', 'requirements.txt'])    
    user_json = deepcopy(USER_JSON_DEFAULT)
    user_json['experiment'] = {'experiment_id': 'old_experiment_id', 'variant': 'A'}
    with open(USER_JSON_PATH, 'w+') as f:
        f.write(json.dumps(user_json))

    venv.run_python_module_command(['mitoinstaller', 'install'])
    current_experiment = get_current_experiment()
    assert current_experiment['experiment_id'] != 'old_experiment_id' and current_experiment['experiment_id'] == get_new_experiment()['experiment_id']

def test_experiment_not_overwritten_by_mitosheet_import(venv: VirtualEnvironment, clear_user_json):
    venv.run_python_module_command(['pip', 'install', '-r', 'requirements.txt'])    
    venv.run_python_module_command(['mitoinstaller', 'install'])
    current_experiment = get_current_experiment()

    venv.run_python_script('import mitosheet')
    current_experiment_new = get_current_experiment()
    assert current_experiment['experiment_id'] == current_experiment_new['experiment_id']
    assert current_experiment['variant'] == current_experiment_new['variant']




import importlib
import os
import sys
from mitoinstaller.commands import run_command

from mitoinstaller.log_utils import log

# The name of the env variable we set
PREFERED_JUPYTER_ENV_VARIABLE_NAME = 'MITO_PREFFERED_JUPYTER'

# The names of which jupyter is prefered
JUPYTER_LAB = 'lab'
JUPYTER_NOTEBOOK = 'notebook'

def is_jupyter_notebook_running() -> bool:
    """
    Format of command output: 

    Currently running servers:
    http://localhost:8888/?token=wenflwenflqnwdlkqmwldkm :: Path/to/launch
    """
    try:
        return '::' in run_command([sys.executable, "-m", 'jupyter', 'notebook', 'list'])[0]
    except:
        return False

def set_prefered_jupyter_env_variable():
    notebook = importlib.util.find_spec('notebook')
    jupyterlab = importlib.util.find_spec('jupyterlab')

    reason = 'default'

    if notebook is None and jupyterlab is None:
        # If the user has nothing installed, we default to lab
        # TODO: in the future, we can randomize here - we just have to log it
        prefered_jupyter = JUPYTER_LAB
        reason = 'neither notebook or lab installed'

    if notebook is not None and jupyterlab is None:
        prefered_jupyter = JUPYTER_NOTEBOOK
        reason = 'only notebook installed'

    elif notebook is None and jupyterlab is not None:
        prefered_jupyter = JUPYTER_LAB
        reason = 'only lab installed'

    else:
        # In the case that the user has both notebook and lab installed, we
        # check if the user has any notebooks running. If they do, we use
        # notebooks. Otherwise, we default to lab
        if is_jupyter_notebook_running():
            prefered_jupyter = JUPYTER_NOTEBOOK
            reason = 'both notebook and lab installed and notebook running'
        else:
            prefered_jupyter = JUPYTER_LAB
            reason = 'both notebook and lab installed and notebook not running'

    os.environ[PREFERED_JUPYTER_ENV_VARIABLE_NAME] = prefered_jupyter
    
    log('setting_prefered_jupyterlab', {
        'prefered_jupyter': prefered_jupyter,
        'prefered_jupyter_reason': reason
    })

def get_prefered_jupyter_env_variable() -> str:
    if PREFERED_JUPYTER_ENV_VARIABLE_NAME not in os.environ or os.environ[PREFERED_JUPYTER_ENV_VARIABLE_NAME] is None:
        # If for some reason the variable is not set, then default to lab
        return JUPYTER_LAB

    return os.environ[PREFERED_JUPYTER_ENV_VARIABLE_NAME]